@@ -19,9 +19,13 @@
 #define LIBPLACEBO_COLORSPACE_H_
 
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
+
 #include <libplacebo/common.h>
 
+PL_API_BEGIN
+
 // The underlying color representation (e.g. RGB, XYZ or YCbCr)
 enum pl_color_system {
     PL_COLOR_SYSTEM_UNKNOWN = 0,
@@ -44,7 +48,7 @@ bool pl_color_system_is_ycbcr_like(enum pl_color_system sys);
 
 // Returns true for color systems that are linear transformations of the RGB
 // equivalent, i.e. are simple matrix multiplications. For color systems with
-// this property, pl_get_decoding_matrix is sufficient for conversion to RGB.
+// this property, `pl_color_repr_decode` is sufficient for conversion to RGB.
 bool pl_color_system_is_linear(enum pl_color_system sys);
 
 // Guesses the best YCbCr-like colorspace based on a image given resolution.
@@ -74,9 +78,13 @@ enum pl_channel {
 // The numerical range of the representation (where applicable).
 enum pl_color_levels {
     PL_COLOR_LEVELS_UNKNOWN = 0,
-    PL_COLOR_LEVELS_TV,         // TV range, e.g. 16-235
-    PL_COLOR_LEVELS_PC,         // PC range, e.g. 0-255
+    PL_COLOR_LEVELS_LIMITED,    // Limited/TV range, e.g. 16-235
+    PL_COLOR_LEVELS_FULL,       // Full/PC range, e.g. 0-255
     PL_COLOR_LEVELS_COUNT,
+
+    // Compatibility aliases
+    PL_COLOR_LEVELS_TV = PL_COLOR_LEVELS_LIMITED,
+    PL_COLOR_LEVELS_PC = PL_COLOR_LEVELS_FULL,
 };
 
 // The alpha representation mode.
@@ -84,6 +92,7 @@ enum pl_alpha_mode {
     PL_ALPHA_UNKNOWN = 0,   // or no alpha channel present
     PL_ALPHA_INDEPENDENT,   // alpha channel is separate from the video
     PL_ALPHA_PREMULTIPLIED, // alpha channel is multiplied into the colors
+    PL_ALPHA_MODE_COUNT,
 };
 
 // The underlying bit-wise representation of a color sample. For example,
@@ -137,7 +146,7 @@ bool pl_color_repr_equal(const struct pl_color_repr *c1,
 
 // Replaces unknown values in the first struct by those of the second struct.
 void pl_color_repr_merge(struct pl_color_repr *orig,
-                         const struct pl_color_repr *new);
+                         const struct pl_color_repr *update);
 
 // This function normalizes the color representation such that
 // color_depth=sample_depth and bit_shift=0; and returns the scaling factor
@@ -146,6 +155,11 @@ void pl_color_repr_merge(struct pl_color_repr *orig,
 // depth will both be inferred as 8 bits for the purposes of this conversion.
 float pl_color_repr_normalize(struct pl_color_repr *repr);
 
+// Guesses the best color levels based on the specified color levels and
+// falling back to using the color system instead. YCbCr-like systems are
+// assumed to be TV range, otherwise this defaults to PC range.
+enum pl_color_levels pl_color_levels_guess(const struct pl_color_repr *repr);
+
 // The colorspace's primaries (gamut)
 enum pl_color_primaries {
     PL_COLOR_PRIM_UNKNOWN = 0,
@@ -154,6 +168,7 @@ enum pl_color_primaries {
     PL_COLOR_PRIM_BT_601_625,   // ITU-R Rec. BT.601 (625-line = PAL, SECAM)
     PL_COLOR_PRIM_BT_709,       // ITU-R Rec. BT.709 (HD), also sRGB
     PL_COLOR_PRIM_BT_470M,      // ITU-R Rec. BT.470 M
+    PL_COLOR_PRIM_EBU_3213,     // EBU Tech. 3213-E / JEDEC P22 phosphors
     // Wide gamut:
     PL_COLOR_PRIM_BT_2020,      // ITU-R Rec. BT.2020 (UltraHD)
     PL_COLOR_PRIM_APPLE,        // Apple RGB
@@ -164,6 +179,7 @@ enum pl_color_primaries {
     PL_COLOR_PRIM_DISPLAY_P3,   // DCI-P3 (Digital Cinema) with D65 white point
     PL_COLOR_PRIM_V_GAMUT,      // Panasonic V-Gamut (VARICAM)
     PL_COLOR_PRIM_S_GAMUT,      // Sony S-Gamut
+    PL_COLOR_PRIM_FILM_C,       // Traditional film primaries with Illuminant C
     PL_COLOR_PRIM_COUNT
 };
 
@@ -182,7 +198,10 @@ enum pl_color_transfer {
     PL_COLOR_TRC_SRGB,          // IEC 61966-2-4 sRGB (CRT emulation)
     PL_COLOR_TRC_LINEAR,        // Linear light content
     PL_COLOR_TRC_GAMMA18,       // Pure power gamma 1.8
+    PL_COLOR_TRC_GAMMA20,       // Pure power gamma 2.0
     PL_COLOR_TRC_GAMMA22,       // Pure power gamma 2.2
+    PL_COLOR_TRC_GAMMA24,       // Pure power gamma 2.4
+    PL_COLOR_TRC_GAMMA26,       // Pure power gamma 2.6
     PL_COLOR_TRC_GAMMA28,       // Pure power gamma 2.8
     PL_COLOR_TRC_PRO_PHOTO,     // ProPhoto RGB (ROMM)
     // High dynamic range:
@@ -200,9 +219,10 @@ enum pl_color_transfer {
 //
 // Note: This returns the highest encodable signal by definition of the EOTF,
 // regardless of the ultimate representation (e.g. scene or display referred).
-// For HLG in particular, this is always 12.0 - which is potentially different
-// from the signal peak after applying the OOTF to go from scene referred to
-// display referred.
+// For HLG in particular, this is always around 3.77 - which is potentially
+// different from the signal peak after applying the OOTF to go from scene
+// referred to display referred (resulting in a display-referred peak of around
+// 4.92 for a 1000 cd/m^2 HLG reference display).
 float pl_color_transfer_nominal_peak(enum pl_color_transfer trc);
 
 static inline bool pl_color_transfer_is_hdr(enum pl_color_transfer trc)
@@ -210,12 +230,20 @@ static inline bool pl_color_transfer_is_hdr(enum pl_color_transfer trc)
     return pl_color_transfer_nominal_peak(trc) > 1.0;
 }
 
-// This defines the standard reference white level (in cd/m^2) that is assumed
-// throughout standards such as those from by ITU-R, EBU, etc.
-// This is particularly relevant for HDR conversions, as this value is used
-// as a reference for conversions between absolute transfer curves (e.g. PQ)
-// and relative transfer curves (e.g. SDR, HLG).
-#define PL_COLOR_REF_WHITE 100.0
+// This defines the display-space standard reference white level (in cd/m^2)
+// that is assumed for SDR content, for use when mapping between HDR and SDR in
+// display space. See ITU-R Report BT.2408 for more information.
+#define PL_COLOR_SDR_WHITE 203.0
+
+// For HLG, which is scene-referred and dependent on the peak luminance of the
+// display device, rather than targeting a fixed cd/m^2 level in display space,
+// we target the 75% level in scene space. This maps to the same brightness
+// level in display space when viewed under the OOTF of a 1000 cd/m^2 HLG
+// reference display.
+#define PL_COLOR_SDR_WHITE_HLG 3.17955
+
+// Compatibility alias for older versions of libplacebo
+#define PL_COLOR_REF_WHITE PL_COLOR_SDR_WHITE
 
 // The semantic interpretation of the decoded image, how is it mastered?
 enum pl_color_light {
@@ -247,13 +275,29 @@ struct pl_color_space {
     enum pl_color_light light;
 
     // The highest value that occurs in the signal, relative to the reference
-    // white. (0 = unknown)
+    // white, alternatively the brightest color value supported by a given
+    // color space. (0 = unknown)
     float sig_peak;
 
     // The average light level that occurs in the signal, relative to the
-    // reference white. (0 = unknown)
+    // reference white, alternatively the desired average brightness level of a
+    // given color space. (0 = unknown)
     float sig_avg;
 
+    // The signal floor, i.e. black point, relative to the reference white.
+    // This effectively determines the signal's contrast, as well. Note that
+    // for most SDR curves, this can only be 0.0 and will effectively be
+    // ignored if otherwise. It only applies to absolute luminance transfer
+    // functions (such as PQ or S-Log), as well as BT.1886.
+    //
+    // Note that a value of 0.0 exactly is treated effectively as "unknown",
+    // and will be defaulted to a contrast of 1000:1 for BT.1886. To override
+    // this logic (thus turning BT.1886 into a true gamma 2.4 function), either
+    // set the transfer function explicitly to PL_COLOR_TRC_GAMMA24, or set
+    // this value to something sufficiently low like FLT_EPSILON. (Never mind
+    // that such high contrast values are unrealistic anyway)
+    float sig_floor;
+
     // Additional scale factor for the signal's reference white. If this is set
     // to a value higher than 1.0, then it's assumed that the signal's encoded
     // reference white is assumed to be brighter than normal by this factor.
@@ -270,9 +314,14 @@ struct pl_color_space {
 // reference white (1.0), after application of the `sig_scale`.
 bool pl_color_space_is_hdr(struct pl_color_space csp);
 
+// Returns whether or not a color space is "black scaled", in which case 0.0 is
+// the true black point. This is true for SDR signals other than BT.1886, as
+// well as for HLG.
+bool pl_color_space_is_black_scaled(struct pl_color_space csp);
+
 // Replaces unknown values in the first struct by those of the second struct.
 void pl_color_space_merge(struct pl_color_space *orig,
-                          const struct pl_color_space *new);
+                          const struct pl_color_space *update);
 
 // Returns whether two colorspaces are exactly identical.
 bool pl_color_space_equal(const struct pl_color_space *c1,
@@ -283,6 +332,13 @@ bool pl_color_space_equal(const struct pl_color_space *c1,
 // PL_COLOR_*_UNKNOWN or 0.0.
 void pl_color_space_infer(struct pl_color_space *space);
 
+// Like `pl_color_space_infer`, but takes default values from the reference
+// color space (excluding certain special cases like HDR or wide gamut). This
+// is basically the logic used by `pl_shader_color_map` to decide the output
+// color space in a conservative way.
+void pl_color_space_infer_ref(struct pl_color_space *space,
+                              const struct pl_color_space *ref);
+
 // Some common color spaces. Note: These don't necessarily have all fields
 // filled, in particular `sig_peak` and `sig_avg` are left unset.
 extern const struct pl_color_space pl_color_space_unknown;
@@ -306,6 +362,8 @@ struct pl_color_adjustment {
     float hue;
     // Gamma adjustment. 1.0 = neutral, 0.0 = solid black
     float gamma;
+    // Color temperature shift. 0.0 = 6500 K, -1.0 = 3000 K, 1.0 = 10000 K
+    float temperature;
 };
 
 // A struct pre-filled with all-neutral values.
@@ -326,6 +384,8 @@ enum pl_chroma_location {
 
 // Fills *x and *y with the offset in luma pixels corresponding to a given
 // chroma location.
+//
+// Note: PL_CHROMA_UNKNOWN defaults to PL_CHROMA_LEFT
 void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y);
 
 // Represents a single CIE xy coordinate (e.g. CIE Yxy with Y = 1.0)
@@ -343,11 +403,27 @@ static inline float pl_cie_Z(struct pl_cie_xy xy) {
     return (1 - xy.x - xy.y) / xy.y;
 }
 
+static inline float pl_cie_xy_equal(const struct pl_cie_xy *a,
+                                    const struct pl_cie_xy *b)
+{
+    return a->x == b->x && a->y == b->y;
+}
+
+// Computes the CIE xy chromaticity coordinates of a CIE D-series illuminant
+// with the given correlated color temperature.
+//
+// `temperature` must be between 2500 K and 25000 K, inclusive.
+struct pl_cie_xy pl_white_from_temp(float temperature);
+
 // Represents the raw physical primaries corresponding to a color space.
 struct pl_raw_primaries {
     struct pl_cie_xy red, green, blue, white;
 };
 
+// Returns whether two raw primaries are exactly identical.
+bool pl_raw_primaries_equal(const struct pl_raw_primaries *a,
+                            const struct pl_raw_primaries *b);
+
 // Returns the raw primaries for a given color space.
 const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim);
 
@@ -367,6 +443,15 @@ struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *s
                                                 const struct pl_raw_primaries *dst,
                                                 enum pl_rendering_intent intent);
 
+// Return a chromatic adaptation matrix, which converts from one white point to
+// another, using the Bradford matrix. This is an RGB->RGB transformation.
+struct pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie_xy dst);
+
+// Returns true if 'b' is entirely contained in 'a'. Useful for figuring out if
+// colorimetric clipping will occur or not.
+bool pl_primaries_superset(const struct pl_raw_primaries *a,
+                           const struct pl_raw_primaries *b);
+
 // Cone types involved in human vision
 enum pl_cone {
     PL_CONE_L = 1 << 0,
@@ -412,15 +497,12 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
                                        const struct pl_raw_primaries *prim);
 
 // Returns a color decoding matrix for a given combination of source color
-// representation and adjustment parameters. This mutates the color_repr to
-// reflect the change. If `params` is left as NULL, it defaults to
-// &pl_color_adjustment_neutral.
+// representation and adjustment parameters. This mutates `repr` to reflect the
+// change. If `params` is NULL, it defaults to &pl_color_adjustment_neutral.
 //
-// This function always performs a conversion to RGB; conversions from
-// arbitrary color representations to other arbitrary other color
-// representations are currently not supported. Not all color systems support
-// all of the color adjustment parameters. (In particular, hue/sat adjustments
-// are currently only supported for YCbCr-like color systems)
+// This function always performs a conversion to RGB. To convert to other
+// colorspaces (e.g. between YUV systems), obtain a second YUV->RGB matrix
+// and invert it using `pl_transform3x3_invert`.
 //
 // Note: For BT.2020 constant-luminance, this outputs chroma information in the
 // range [-0.5, 0.5]. Since the CL system conversion is non-linear, further
@@ -445,7 +527,9 @@ struct pl_icc_profile {
     // If a profile is set, this signature must uniquely identify it. It could
     // be, for example, a checksum of the profile contents. Alternatively, it
     // could be the pointer to the ICC profile itself, as long as the user
-    // makes sure that this memory is used in an immutable way.
+    // makes sure that this memory is used in an immutable way. For a third
+    // possible interpretation, consider simply incrementing this uint64_t
+    // every time you suspect the profile has changed.
     uint64_t signature;
 };
 
@@ -453,4 +537,11 @@ struct pl_icc_profile {
 bool pl_icc_profile_equal(const struct pl_icc_profile *p1,
                           const struct pl_icc_profile *p2);
 
+// Sets `signature` to a hash of `profile->data`, if non-NULL. Provided as a
+// convenience function for the sake of users ingesting arbitrary ICC profiles
+// from sources where they can't reliably detect profile changes.
+void pl_icc_profile_compute_signature(struct pl_icc_profile *profile);
+
+PL_API_END
+
 #endif // LIBPLACEBO_COLORSPACE_H_
