@@ -19,11 +19,11 @@
 #define LIBPLACEBO_CONFIG_H_
 
 // Increased any time the library changes in a fundamental/major way.
-#define PL_MAJOR_VER 3
+#define PL_MAJOR_VER 4
 
 // Increased any time the API changes. (Note: Does not reset when PL_MAJOR_VER
 // is increased)
-#define PL_API_VER 120
+#define PL_API_VER 157
 
 // Increased any time a fix is made to a given API version.
 #define PL_FIX_VER (pl_fix_ver())
@@ -42,9 +42,14 @@ const char *pl_version(void);
 #undef PL_HAVE_SHADERC
 #undef PL_HAVE_VULKAN
 #define PL_HAVE_OPENGL 1
+#undef PL_HAVE_D3D11
 
 
 // Extra compiler-specific stuff
+#if defined(_MSC_VER)
+#define PL_DEPRECATED
+#else
 #define PL_DEPRECATED __attribute__((deprecated))
+#endif
 
 #endif // LIBPLACEBO_CONTEXT_H_
