@@ -24,27 +24,30 @@
 #include <libplacebo/gpu.h>
 #include <libplacebo/shaders/av1.h>
 #include <libplacebo/shaders/colorspace.h>
+#include <libplacebo/shaders/lut.h>
 #include <libplacebo/shaders/sampling.h>
 #include <libplacebo/shaders/custom.h>
 #include <libplacebo/swapchain.h>
 
-struct pl_renderer;
+PL_API_BEGIN
+
+// Thread-safety: Unsafe
+typedef PL_STRUCT(pl_renderer) *pl_renderer;
 
 // Creates a new renderer object, which is backed by a GPU context. This is a
 // high-level object that takes care of the rendering chain as a whole, from
 // the source textures to the finished frame.
-struct pl_renderer *pl_renderer_create(struct pl_context *ctx,
-                                       const struct pl_gpu *gpu);
-void pl_renderer_destroy(struct pl_renderer **rr);
+pl_renderer pl_renderer_create(pl_log log, pl_gpu gpu);
+void pl_renderer_destroy(pl_renderer *rr);
 
 // Saves the internal shader cache of this renderer into an abstract cache
 // object that can be saved to disk and later re-loaded to speed up
 // recompilation of shaders. See `pl_dispatch_save` for more information.
-size_t pl_renderer_save(struct pl_renderer *rr, uint8_t *out_cache);
+size_t pl_renderer_save(pl_renderer rr, uint8_t *out_cache);
 
 // Load the result of a previous `pl_renderer_save` call. See
 // `pl_dispatch_load` for more information.
-void pl_renderer_load(struct pl_renderer *rr, const uint8_t *cache);
+void pl_renderer_load(pl_renderer rr, const uint8_t *cache);
 
 enum pl_lut_type {
     PL_LUT_UNKNOWN = 0,
@@ -64,6 +67,24 @@ enum pl_lut_type {
     // LUT's tagged metadata, and otherwise falls back to PL_LUT_NATIVE.
 };
 
+enum pl_render_stage {
+    PL_RENDER_STAGE_FRAME,  // full frame redraws, for fresh/uncached frames
+    PL_RENDER_STAGE_BLEND,  // the output blend pass (only for pl_render_image_mix)
+    PL_RENDER_STAGE_COUNT,
+};
+
+struct pl_render_info {
+    const struct pl_dispatch_info *pass;    // information about the shader
+    enum pl_render_stage stage;             // the associated render stage
+
+    // For PL_RENDER_STAGE_FRAME, this specifies the chronological index
+    // of this pass within the frame (starting at `index == 0`).
+    //
+    // For PL_RENDER_STAGE_BLEND, this specifies the number of frames
+    // being blended (since that results in a different shader).
+    int index;
+};
+
 // Represents the options used for rendering. These affect the quality of
 // the result.
 struct pl_render_params {
@@ -93,10 +114,6 @@ struct pl_render_params {
     // this must be a filter config with `polar` set to false, since it's only
     // used for 1D mixing and thus only 1D filters are compatible.
     //
-    // As a special case, if `frame_mixer->kernel` is NULL, then libplacebo
-    // will use a built-in, inexpensive and relatively unobtrusive oversampling
-    // frame mixing algorithm. (See `pl_oversample_frame_mixer`)
-    //
     // If set to NULL, frame mixing is disabled, in which case
     // `pl_render_image_mix` behaves as `pl_render_image`, also completely
     // bypassing the mixing cache.
@@ -164,6 +181,13 @@ struct pl_render_params {
     const struct pl_custom_lut *lut;
     enum pl_lut_type lut_type;
 
+    // If the image being rendered does not span the entire size of the target,
+    // it will be cleared explicitly using this background color (RGB). To
+    // disable this logic, set `skip_target_clearing`.
+    float background_color[3];
+    float background_transparency; // 0.0 for opaque, 1.0 for fully transparent
+    bool skip_target_clearing;
+
     // --- Performance / quality trade-off options:
     // These should generally be left off where quality is desired, as they can
     // degrade the result quite noticeably; but may be useful for older or
@@ -182,11 +206,6 @@ struct pl_render_params {
     // `pl_sample_filter_params` for more information.
     float polar_cutoff;
 
-    // Skips dispatching the high-quality scalers for overlay textures, and
-    // always falls back to built-in GPU samplers. Note: The scalers are
-    // already disabled if the overlay texture does not need to be scaled.
-    bool disable_overlay_sampling;
-
     // Allows the peak detection result to be delayed by up to a single frame,
     // which can sometimes (not always) allow skipping some otherwise redundant
     // sampling work. Only relevant when peak detection is active (i.e.
@@ -222,6 +241,10 @@ struct pl_render_params {
     // reduction step, if one is performed.
     bool force_icc_lut;
 
+    // Ignore ICC profiles attached to either `image` or `target`.
+    // Note: A LUT may still be generated if `force_icc_lut` is also enabled.
+    bool ignore_icc_profiles;
+
     // Forces the use of dithering, even when rendering to 16-bit FBOs. This is
     // generally pretty pointless because most 16-bit FBOs have high enough
     // depth that rounding errors are below the human perception threshold,
@@ -232,9 +255,29 @@ struct pl_render_params {
     // texture format available. This disables most features.
     bool disable_fbos;
 
+    // If this is true, all shaders will be generated as "dynamic" shaders,
+    // with any compile-time constants being replaced by runtime-adjustable
+    // values. This is generally a performance loss, but has the advantage of
+    // being able to freely change parameters without triggering shader
+    // recompilations.
+    //
+    // It's a good idea to enable while presenting configurable settings to the
+    // user, but it should be set to false once those values are "dialed in".
+    bool dynamic_constants;
+
+    // This callback is invoked for every pass successfully executed in the
+    // process of rendering a frame. Optional.
+    //
+    // Note: `info` is only valid until this function returns.
+    void (*info_callback)(void *priv, const struct pl_render_info *info);
+    void *info_priv;
+
     // --- Deprecated aliases
     const struct pl_icc_params *lut3d_params PL_DEPRECATED; // fallback for `icc_params`
     bool force_3dlut PL_DEPRECATED; // fallback for `force_icc_lut`
+
+    // --- Deprecated/removed fields
+    bool disable_overlay_sampling PL_DEPRECATED; // no longer used
 };
 
 // This contains the default/recommended options for reasonable image quality,
@@ -254,14 +297,25 @@ extern const struct pl_render_params pl_render_default_params;
 // and where maximum image quality is desired.
 extern const struct pl_render_params pl_render_high_quality_params;
 
-// Special filter config for the built-in oversampling frame mixing algorithm.
-// This is equivalent to (struct pl_filter_config) {0}.
-extern const struct pl_filter_config pl_oversample_frame_mixer;
+// Special filter config for the built-in oversampling algorithm. This is an
+// opaque filter with no meaningful representation. though it has one tunable
+// parameter controlling the threshold at which to switch back to ordinary
+// nearest neighbour sampling. (See `pl_shader_sample_oversample`)
+extern const struct pl_filter_config pl_filter_oversample;
+
+// Backwards compatibility
+#define pl_oversample_frame_mixer pl_filter_oversample
 
 // A list of recommended frame mixer presets, terminated by {0}
 extern const struct pl_filter_preset pl_frame_mixers[];
 extern const int pl_num_frame_mixers; // excluding trailing {0}
 
+// A list of recommended scaler presets, terminated by {0}. This is almost
+// equivalent to `pl_filter_presets` with the exception of including extra
+// built-in filters that don't map to the `pl_filter` architecture.
+extern const struct pl_filter_preset pl_scale_filters[];
+extern const int pl_num_scale_filters; // excluding trailing {0}
+
 #define PL_MAX_PLANES 4
 
 // High level description of a single slice of an image. This basically
@@ -270,7 +324,7 @@ struct pl_plane {
     // The texture underlying this plane. The texture must be 2D, and must
     // have specific parameters set depending on what the plane is being used
     // for (see `pl_render_image`).
-    const struct pl_tex *texture;
+    pl_tex texture;
 
     // The preferred behaviour when sampling outside of this texture. Optional,
     // since the default (PL_TEX_ADDRESS_CLAMP) is very reasonable.
@@ -328,34 +382,41 @@ enum pl_overlay_mode {
     PL_OVERLAY_MODE_COUNT,
 };
 
+struct pl_overlay_part {
+    struct pl_rect2df src; // source coordinate with respect to `tex`
+    struct pl_rect2d dst;  // target coordinates with respect to the frame
+
+    // If `mode` is PL_OVERLAY_MONOCHROME, then this specifies the color of
+    // this overlay part. The color is multiplied into the sampled texture's
+    // first channel.
+    float color[4];
+};
+
 // A struct representing an image overlay (e.g. for subtitles or on-screen
 // status messages, controls, ...)
 struct pl_overlay {
-    // The plane to overlay. Multi-plane overlays are not supported. If
-    // necessary, multiple planes can be combined by treating them as separate
-    // overlays with different base colors.
-    //
-    // Must have `params.sampleable` set, and it's recommended to also have the
-    // sample mode set to `PL_TEX_SAMPLE_LINEAR`.
-    //
-    // Note: shift_x/y are simply treated as a uniform sampling offset.
-    struct pl_plane plane;
-
-    // The (absolute) coordinates at which to render this overlay texture. May
-    // be flipped, and partially or wholly outside the image. If the size does
-    // not exactly match the texture, it will be scaled/stretched to fit.
-    struct pl_rect2d rect;
+    // The texture containing the backing data for overlay parts. Must have
+    // `params.sampleable` set.
+    pl_tex tex;
 
     // This controls the coloring mode of this overlay.
     enum pl_overlay_mode mode;
-    // If `mode` is PL_OVERLAY_MONOCHROME, then the texture is treated as an
-    // alpha map and multiplied by this base color. Ignored for the other modes.
-    float base_color[3];
 
     // This controls the colorspace information for this overlay. The contents
     // of the texture / the value of `color` are interpreted according to this.
     struct pl_color_repr repr;
     struct pl_color_space color;
+
+    // The number of parts for this overlay.
+    const struct pl_overlay_part *parts;
+    int num_parts;
+
+    // (Deprecated) These fields exist for backwards compatibility. They must
+    // not be used as the same times as `tex`. They are interpreted as an
+    // overlay with a single part.
+    struct pl_plane plane PL_DEPRECATED;
+    struct pl_rect2d rect PL_DEPRECATED; // analog to `pl_overlay_part.dst`
+    float base_color[3] PL_DEPRECATED; // analog to `pl_overlay_part.color`
 };
 
 // High-level description of a complete frame, including metadata and planes
@@ -395,7 +456,7 @@ struct pl_frame {
     int num_overlays;
 
     // Note on subsampling and plane correspondence: All planes belonging to
-    // the same frame will only be streched by an integer multiple (or inverse
+    // the same frame will only be stretched by an integer multiple (or inverse
     // thereof) in order to match the reference dimensions of this image. For
     // example, suppose you have an 8x4 image. A valid plane scaling would be
     // 4x2 -> 8x4 or 4x4 -> 4x4, but not 6x4 -> 8x4. So if a 6x4 plane is
@@ -417,12 +478,15 @@ struct pl_frame {
     // un-applying grain makes little sense.
     struct pl_av1_grain_data av1_grain;
 
+    // Ignored by libplacebo. May be useful for users.
+    void *user_data;
+
     // Deprecated fields provided merely for backwards compatibility. The
     // use of these should be discontinued as soon as possible.
     int width PL_DEPRECATED; // ignored
     int height PL_DEPRECATED;
     uint64_t signature PL_DEPRECATED; // ignored
-    const struct pl_tex *fbo PL_DEPRECATED; // fallback for `target.planes`
+    pl_tex fbo PL_DEPRECATED; // fallback for `target.planes`
     struct pl_rect2df src_rect PL_DEPRECATED; // fallback for `image.crop`
     struct pl_rect2df dst_rect PL_DEPRECATED; // fallback for `target.crop`
 };
@@ -444,12 +508,17 @@ bool pl_frame_is_cropped(const struct pl_frame *frame);
 
 // Helper function to reset a frame to a given RGB color. If the frame's
 // color representation is something other than RGB, the clear color will
-// be adjusted accordingly.
-//
-// Note: For sake of simplicity, this function only works with opaque clears,
-// i.e. A=1. If you need to clear to transparent colors, use `pl_tex_clear`.
-void pl_frame_clear(const struct pl_gpu *gpu, const struct pl_frame *frame,
-                    const float clear_color[3]);
+// be adjusted accordingly. `clear_color` should be non-premultiplied.
+void pl_frame_clear_rgba(pl_gpu gpu, const struct pl_frame *frame,
+                         const float clear_color[4]);
+
+// Like `pl_frame_clear_rgba` but without an alpha channel.
+static inline void pl_frame_clear(pl_gpu gpu, const struct pl_frame *frame,
+                                  const float clear_color[3])
+{
+    const float clear_color_rgba[4] = { clear_color[0], clear_color[1], clear_color[2], 1.0 };
+    pl_frame_clear_rgba(gpu, frame, clear_color_rgba);
+}
 
 // Deprecated aliases, provided for backwards compatibility
 #define pl_image pl_frame
@@ -501,7 +570,10 @@ static PL_DEPRECATED inline bool pl_render_target_partial(
 // which means they get affected by things like scaling and frame mixing.
 // `target.overlays` will also be rendered, but directly onto the target. They
 // don't even need to be inside `target.crop`.
-bool pl_render_image(struct pl_renderer *rr, const struct pl_frame *image,
+//
+// Note: `image` may be NULL, in which case `target.overlays` will still be
+// rendered, but nothing else.
+bool pl_render_image(pl_renderer rr, const struct pl_frame *image,
                      const struct pl_frame *target,
                      const struct pl_render_params *params);
 
@@ -512,7 +584,7 @@ bool pl_render_image(struct pl_renderer *rr, const struct pl_frame *image,
 // purge some state related to things like HDR peak detection or frame mixing,
 // so calling it is a good idea if the content source is expected to change
 // dramatically (e.g. when switching to a different file).
-void pl_renderer_flush_cache(struct pl_renderer *rr);
+void pl_renderer_flush_cache(pl_renderer rr);
 
 // Represents a mixture of input frames, distributed temporally.
 //
@@ -523,7 +595,9 @@ struct pl_frame_mix {
     // sufficient to meet the needs of the configured frame mixer. See the
     // section below for more information.
     //
-    // At least one frame is always required, i.e. `num_frames > 0`.
+    // If the number of frames is 0, this call will be equivalent to
+    // `pl_render_image` with `image == NULL`. If the number of frames is 1,
+    // this is equivalent to `pl_render_image` with that image.
     int num_frames;
 
     // A list of the frames themselves. The frames can have different
@@ -584,9 +658,11 @@ struct pl_frame_mix {
 // Helper function to calculate the frame mixing radius.
 static inline float pl_frame_mix_radius(const struct pl_render_params *params)
 {
-    return (params->frame_mixer && params->frame_mixer->kernel)
-        ? params->frame_mixer->kernel->radius
-        : 0.0;
+    // For backwards compatibility, allow !frame_mixer->kernel
+    if (!params->frame_mixer || !params->frame_mixer->kernel)
+        return 0.0;
+
+    return params->frame_mixer->kernel->radius;
 }
 
 // Render a mixture of images to the target using the given parameters. This
@@ -597,8 +673,10 @@ static inline float pl_frame_mix_radius(const struct pl_render_params *params)
 // This allows libplacebo to perform rudimentary frame mixing / interpolation,
 // in order to eliminate judder artifacts typically associated with
 // source/display frame rate mismatch.
-bool pl_render_image_mix(struct pl_renderer *rr, const struct pl_frame_mix *images,
+bool pl_render_image_mix(pl_renderer rr, const struct pl_frame_mix *images,
                          const struct pl_frame *target,
                          const struct pl_render_params *params);
 
+PL_API_END
+
 #endif // LIBPLACEBO_RENDERER_H_
