@@ -24,6 +24,8 @@
 #include <libplacebo/colorspace.h>
 #include <libplacebo/shaders.h>
 
+PL_API_BEGIN
+
 // Decode the color into normalized RGB, given a specified color_repr. This
 // also takes care of additional pre- and post-conversions requires for the
 // "special" color systems (XYZ, BT.2020-C, etc.). If `params` is left as NULL,
@@ -31,28 +33,25 @@
 //
 // Note: This function always returns PC-range RGB with pre-multiplied alpha.
 // It mutates the pl_color_repr to reflect the change.
-void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
+void pl_shader_decode_color(pl_shader sh, struct pl_color_repr *repr,
                             const struct pl_color_adjustment *params);
 
 // Encodes a color from normalized, PC-range, pre-multiplied RGB into a given
 // representation. That is, this performs the inverse operation of
 // `pl_shader_decode_color` (sans color adjustments).
-void pl_shader_encode_color(struct pl_shader *sh,
-                            const struct pl_color_repr *repr);
+void pl_shader_encode_color(pl_shader sh, const struct pl_color_repr *repr);
 
-// Linearize (expand) `vec4 color`, given a specified color_transfer. In
-// essence, this loosely corresponds to the ITU-R EOTF, calculated on an
-// idealized (reference) monitor with a white point of PL_COLOR_SDR_WHITE and
-// infinite contrast.
+// Linearize (expand) `vec4 color`, given a specified color space. In essence,
+// this corresponds to the ITU-R EOTF.
 //
 // Note: Unlike the ITU-R EOTF, it never includes the OOTF - even for systems
 // where the EOTF includes the OOTF (such as HLG).
-void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc);
+void pl_shader_linearize(pl_shader sh, struct pl_color_space csp);
 
-// Delinearize (compress), given a TRC as output. This loosely corresponds to
-// the inverse EOTF (not the OETF) in ITU-R terminology, again assuming a
-// reference monitor.
-void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc);
+// Delinearize (compress), given a color space as output. This loosely
+// corresponds to the inverse EOTF (not the OETF) in ITU-R terminology, again
+// assuming a reference monitor.
+void pl_shader_delinearize(pl_shader sh, struct pl_color_space csp);
 
 struct pl_sigmoid_params {
     // The center (bias) of the sigmoid curve. Must be between 0.0 and 1.0.
@@ -73,12 +72,10 @@ extern const struct pl_sigmoid_params pl_sigmoid_default_params;
 //
 // Warning: This function clamps the input to the interval [0,1]; and as such
 // it should *NOT* be used on already-decoded high-dynamic range content.
-void pl_shader_sigmoidize(struct pl_shader *sh,
-                          const struct pl_sigmoid_params *params);
+void pl_shader_sigmoidize(pl_shader sh, const struct pl_sigmoid_params *params);
 
 // This performs the inverse operation to `pl_shader_sigmoidize`.
-void pl_shader_unsigmoidize(struct pl_shader *sh,
-                            const struct pl_sigmoid_params *params);
+void pl_shader_unsigmoidize(pl_shader sh, const struct pl_sigmoid_params *params);
 
 struct pl_peak_detect_params {
     // Smoothing coefficient for the detected values. This controls the time
@@ -92,7 +89,7 @@ struct pl_peak_detect_params {
     float smoothing_period;
 
     // In order to avoid reacting sluggishly on scene changes as a result of
-    // the low-pass filter, we disable it when the difference betwen the
+    // the low-pass filter, we disable it when the difference between the
     // current frame brightness and the average frame brightness exceeds a
     // given threshold difference. But rather than a single hard cutoff, which
     // would lead to weird discontinuities on fades, we gradually disable it
@@ -109,6 +106,11 @@ struct pl_peak_detect_params {
     // by this amount, as a percentage of the actual measured peak. If left
     // as 0.0, this logic is disabled. The default value is 0.05.
     float overshoot_margin;
+
+    // To avoid over-tone-mapping very dark scenes (or black frames), this
+    // imposes a hard lower bound on the detected peak. If left as 0.0, it
+    // instead defaults to a value of 1.0.
+    float minimum_peak;
 };
 
 extern const struct pl_peak_detect_params pl_peak_detect_default_params;
@@ -127,9 +129,8 @@ extern const struct pl_peak_detect_params pl_peak_detect_default_params;
 // The parameter `csp` holds the representation of the color values that are
 // the input to this function. (They must already be in decoded RGB form, i.e.
 // alternate color representations are not supported)
-bool pl_shader_detect_peak(struct pl_shader *sh,
-                           struct pl_color_space csp,
-                           struct pl_shader_obj **state,
+bool pl_shader_detect_peak(pl_shader sh, struct pl_color_space csp,
+                           pl_shader_obj *state,
                            const struct pl_peak_detect_params *params);
 
 // After dispatching the above shader, this function *may* be used to read out
@@ -141,7 +142,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
 // function is *not* needed when the user only wants to use `pl_shader_color_map`,
 // since that can ingest the peak detection state object directly. It only
 // serves as a utility/debugging function.
-bool pl_get_detected_peak(const struct pl_shader_obj *state,
+bool pl_get_detected_peak(const pl_shader_obj state,
                           float *out_peak, float *out_avg);
 
 // A collection of various tone mapping algorithms supported by libplacebo.
@@ -195,7 +196,7 @@ enum pl_tone_mapping_algorithm {
     // Linearly stretches the source gamut to the destination gamut. This will
     // preserve all details accurately, but results in a significantly darker
     // image. Best when combined with peak detection. The parameter can be used
-    // as an aditional scaling coefficient to make the image (linearly)
+    // as an additional scaling coefficient to make the image (linearly)
     // brighter or darker. Defaults to 1.0.
     PL_TONE_MAPPING_LINEAR,
 
@@ -235,8 +236,9 @@ struct pl_color_map_params {
     // If you want to always use the saturating/colorimetric tone mapping, set
     // the strength to 0.0. If you want to always use the desaturating/hollywood
     // tone mapping, set the strength to 1.0 and the exponent to 0.0. The
-    // default settings are strength 0.75, exponent 1.5 and base 0.18, which
-    // provides a reasonable balance.
+    // default settings are strength 0.90, exponent 0.2 and base 0.18, which
+    // provide a fairly strong preset while nonetheless preserving some amount
+    // of colorimetric accuracy.
     float desaturation_strength;
     float desaturation_exponent;
     float desaturation_base;
@@ -277,16 +279,16 @@ extern const struct pl_color_map_params pl_color_map_default_params;
 // frame. If frame-level accuracy is desired, then users should call
 // `pl_shader_detect_peak` separately and dispatch the resulting shader
 // *before* dispatching this one.
-void pl_shader_color_map(struct pl_shader *sh,
+void pl_shader_color_map(pl_shader sh,
                          const struct pl_color_map_params *params,
                          struct pl_color_space src, struct pl_color_space dst,
-                         struct pl_shader_obj **peak_detect_state,
+                         pl_shader_obj *peak_detect_state,
                          bool prelinearized);
 
 // Applies a set of cone distortion parameters to `vec4 color` in a given color
 // space. This can be used to simulate color blindness. See `pl_cone_params`
 // for more information.
-void pl_shader_cone_distort(struct pl_shader *sh, struct pl_color_space csp,
+void pl_shader_cone_distort(pl_shader sh, struct pl_color_space csp,
                             const struct pl_cone_params *params);
 
 enum pl_dither_method {
@@ -352,8 +354,10 @@ extern const struct pl_dither_params pl_dither_default_params;
 // of the resulting image. When doing low bit depth dithering for aesthetic
 // purposes, it's recommended that the user explicitly (de)linearize the colors
 // before and after this algorithm.
-void pl_shader_dither(struct pl_shader *sh, int new_depth,
-                      struct pl_shader_obj **dither_state,
+void pl_shader_dither(pl_shader sh, int new_depth,
+                      pl_shader_obj *dither_state,
                       const struct pl_dither_params *params);
 
+PL_API_END
+
 #endif // LIBPLACEBO_SHADERS_COLORSPACE_H_
