@@ -22,6 +22,7 @@
 #include <assert.h>
 
 #include <libavutil/hwcontext.h>
+#include <libavutil/imgutils.h>
 #include <libavutil/mastering_display_metadata.h>
 #include <libavutil/pixdesc.h>
 
@@ -185,7 +186,10 @@ static inline enum AVColorTransferCharacteristic pl_transfer_to_av(enum pl_color
     case PL_COLOR_TRC_SRGB:         return AVCOL_TRC_IEC61966_2_1;
     case PL_COLOR_TRC_LINEAR:       return AVCOL_TRC_LINEAR;
     case PL_COLOR_TRC_GAMMA18:      return AVCOL_TRC_UNSPECIFIED; // missing
+    case PL_COLOR_TRC_GAMMA20:      return AVCOL_TRC_UNSPECIFIED; // missing
     case PL_COLOR_TRC_GAMMA22:      return AVCOL_TRC_GAMMA22;
+    case PL_COLOR_TRC_GAMMA24:      return AVCOL_TRC_UNSPECIFIED; // missing
+    case PL_COLOR_TRC_GAMMA26:      return AVCOL_TRC_UNSPECIFIED; // missing
     case PL_COLOR_TRC_GAMMA28:      return AVCOL_TRC_GAMMA28;
     case PL_COLOR_TRC_PRO_PHOTO:    return AVCOL_TRC_UNSPECIFIED; // missing
     case PL_COLOR_TRC_PQ:           return AVCOL_TRC_SMPTE2084;
@@ -351,7 +355,7 @@ misaligned:
     return planes;
 }
 
-static inline bool pl_test_pixfmt(const struct pl_gpu *gpu,
+static inline bool pl_test_pixfmt(pl_gpu gpu,
                                   enum AVPixelFormat pixfmt)
 {
     struct pl_bit_encoding bits;
@@ -385,7 +389,7 @@ static inline void pl_avframe_set_color(AVFrame *frame, struct pl_color_space sp
 
         if (sd) {
             AVContentLightMetadata *clm = (AVContentLightMetadata *) sd->data;
-            *clm = (struct AVContentLightMetadata) {
+            *clm = (AVContentLightMetadata) {
                 .MaxCLL = space.sig_peak * PL_COLOR_SDR_WHITE,
                 .MaxFALL = space.sig_avg * PL_COLOR_SDR_WHITE,
             };
@@ -395,6 +399,23 @@ static inline void pl_avframe_set_color(AVFrame *frame, struct pl_color_space sp
         }
     }
 
+    if (space.sig_floor > 0.0) {
+        sd = av_frame_get_side_data(frame, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA);
+        if (!sd) {
+            sd = av_frame_new_side_data(frame, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA,
+                                        sizeof(AVMasteringDisplayMetadata));
+        }
+
+        if (sd) {
+            AVMasteringDisplayMetadata *mdm = (AVMasteringDisplayMetadata *) sd->data;
+            *mdm = (AVMasteringDisplayMetadata) {
+                .max_luminance = av_d2q(space.sig_peak * PL_COLOR_SDR_WHITE, 100000),
+                .min_luminance = av_d2q(space.sig_floor * PL_COLOR_SDR_WHITE, 100000),
+                .has_luminance = 1,
+            };
+        }
+    }
+
     // No way to map space.sig_scale, so just ignore it
 }
 
@@ -511,8 +532,10 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
     // This overrides the CLL values above, if both are present
     if ((sd = av_frame_get_side_data(frame, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA))) {
         const AVMasteringDisplayMetadata *mdm = (AVMasteringDisplayMetadata *) sd->data;
-        if (mdm->has_luminance)
+        if (mdm->has_luminance) {
             out->color.sig_peak = av_q2d(mdm->max_luminance) / PL_COLOR_SDR_WHITE;
+            out->color.sig_floor = av_q2d(mdm->min_luminance) / PL_COLOR_SDR_WHITE;
+        }
     }
 
     // Make sure this value is more or less legal
@@ -576,9 +599,47 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
     }
 }
 
-static inline bool pl_frame_recreate_from_avframe(const struct pl_gpu *gpu,
+static inline void pl_swapchain_colors_from_avframe(struct pl_swapchain_colors *out_colors,
+                                                    const AVFrame *avframe)
+{
+    const AVFrameSideData *sd;
+    struct pl_frame frame;
+    pl_frame_from_avframe(&frame, avframe);
+
+    *out_colors = (struct pl_swapchain_colors) {
+        .primaries = frame.color.primaries,
+        .transfer = frame.color.transfer,
+    };
+
+    if ((sd = av_frame_get_side_data(avframe, AV_FRAME_DATA_CONTENT_LIGHT_LEVEL))) {
+        const AVContentLightMetadata *clm = (AVContentLightMetadata *) sd->data;
+        out_colors->hdr.max_cll = clm->MaxCLL;
+        out_colors->hdr.max_fall = clm->MaxFALL;
+    }
+
+    if ((sd = av_frame_get_side_data(avframe, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA))) {
+        const AVMasteringDisplayMetadata *mdm = (AVMasteringDisplayMetadata *) sd->data;
+        if (mdm->has_luminance) {
+            out_colors->hdr.min_luma = av_q2d(mdm->min_luminance);
+            out_colors->hdr.max_luma = av_q2d(mdm->max_luminance);
+        }
+
+        if (mdm->has_primaries) {
+            out_colors->hdr.prim.red.x   = av_q2d(mdm->display_primaries[0][0]);
+            out_colors->hdr.prim.red.y   = av_q2d(mdm->display_primaries[0][1]);
+            out_colors->hdr.prim.green.x = av_q2d(mdm->display_primaries[1][0]);
+            out_colors->hdr.prim.green.y = av_q2d(mdm->display_primaries[1][1]);
+            out_colors->hdr.prim.blue.x  = av_q2d(mdm->display_primaries[2][0]);
+            out_colors->hdr.prim.blue.y  = av_q2d(mdm->display_primaries[2][1]);
+            out_colors->hdr.prim.white.x = av_q2d(mdm->white_point[0]);
+            out_colors->hdr.prim.white.y = av_q2d(mdm->white_point[1]);
+        }
+    }
+}
+
+static inline bool pl_frame_recreate_from_avframe(pl_gpu gpu,
                                                   struct pl_frame *out,
-                                                  const struct pl_tex *tex[4],
+                                                  pl_tex tex[4],
                                                   const AVFrame *frame)
 {
     const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
@@ -592,8 +653,8 @@ static inline bool pl_frame_recreate_from_avframe(const struct pl_gpu *gpu,
 
     for (int p = 0; p < planes; p++) {
         bool is_chroma = p == 1 || p == 2; // matches lavu logic
-        data[p].width = frame->width >> (is_chroma ? desc->log2_chroma_w : 0);
-        data[p].height = frame->height >> (is_chroma ? desc->log2_chroma_h : 0);
+        data[p].width = AV_CEIL_RSHIFT(frame->width, is_chroma ? desc->log2_chroma_w : 0);
+        data[p].height = AV_CEIL_RSHIFT(frame->height, is_chroma ? desc->log2_chroma_h : 0);
 
         if (!pl_recreate_plane(gpu, &out->planes[p], &tex[p], &data[p]))
             return false;
@@ -608,9 +669,18 @@ static void pl_avframe_free(void *priv)
     av_frame_free(&frame);
 }
 
-static inline bool pl_upload_avframe(const struct pl_gpu *gpu,
+#define PL_MAGIC0 0xfb5b3b8b
+#define PL_MAGIC1 0xee659f6d
+
+struct pl_avalloc {
+    uint32_t magic[2];
+    pl_gpu gpu;
+    pl_buf buf;
+};
+
+static inline bool pl_upload_avframe(pl_gpu gpu,
                                      struct pl_frame *out,
-                                     const struct pl_tex *tex[4],
+                                     pl_tex tex[4],
                                      const AVFrame *frame)
 {
     const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
@@ -626,14 +696,26 @@ static inline bool pl_upload_avframe(const struct pl_gpu *gpu,
     if (!planes)
         return false;
 
+    // Probe for frames allocated by our get_buffer2
+    struct pl_avalloc *alloc = frame->buf[0] ? av_buffer_get_opaque(frame->buf[0]) : NULL;
+    pl_buf buf = NULL;
+    if (alloc && alloc->magic[0] == PL_MAGIC0 && alloc->magic[1] == PL_MAGIC1) {
+        assert(alloc->gpu == gpu);
+        buf = alloc->buf;
+    }
+
     for (int p = 0; p < planes; p++) {
         bool is_chroma = p == 1 || p == 2; // matches lavu logic
-        data[p].width = frame->width >> (is_chroma ? desc->log2_chroma_w : 0);
-        data[p].height = frame->height >> (is_chroma ? desc->log2_chroma_h : 0);
+        data[p].width = AV_CEIL_RSHIFT(frame->width, is_chroma ? desc->log2_chroma_w : 0);
+        data[p].height = AV_CEIL_RSHIFT(frame->height, is_chroma ? desc->log2_chroma_h : 0);
         data[p].row_stride = frame->linesize[p];
         data[p].pixels = frame->data[p];
 
-        if (gpu->caps & PL_GPU_CAP_CALLBACKS) {
+        if (buf) {
+            data[p].pixels = NULL;
+            data[p].buf = buf;
+            data[p].buf_offset = (uintptr_t) frame->data[p] - (uintptr_t) buf->data;
+        } else if (gpu->limits.callbacks) {
             // Use asynchronous upload if possible
             data[p].callback = pl_avframe_free;
             data[p].priv = av_frame_clone(frame);
@@ -654,7 +736,7 @@ static void pl_done_cb(void *priv)
     *status = true;
 }
 
-static inline bool pl_download_avframe(const struct pl_gpu *gpu,
+static inline bool pl_download_avframe(pl_gpu gpu,
                                        const struct pl_frame *frame,
                                        AVFrame *out_frame)
 {
@@ -685,4 +767,113 @@ static inline bool pl_download_avframe(const struct pl_gpu *gpu,
     return true;
 }
 
+#define PL_ALIGN(x, align) ((align) ? ((x) + (align) - 1) / (align) * (align) : (x))
+#define PL_ALIGN2(x, align) (((x) + (align) - 1) & ~((align) - 1))
+#define PL_MAX(x, y) ((x) > (y) ? (x) : (y))
+
+static inline void pl_avalloc_free(void *opaque, uint8_t *data)
+{
+    struct pl_avalloc *alloc = opaque;
+    assert(alloc->magic[0] == PL_MAGIC0);
+    assert(alloc->magic[1] == PL_MAGIC1);
+    assert(alloc->buf->data == data);
+    pl_buf_destroy(alloc->gpu, &alloc->buf);
+    free(alloc);
+}
+
+static inline int pl_get_buffer2(AVCodecContext *avctx, AVFrame *pic, int flags)
+{
+    int linesize_align[AV_NUM_DATA_POINTERS];
+    uint8_t *pointers[5];
+    size_t offsets[4];
+    int width = pic->width;
+    int height = pic->height;
+    size_t total_size = 0;
+    int ret;
+
+    pl_gpu *pgpu = avctx->opaque;
+    pl_gpu gpu = pgpu ? *pgpu : NULL;
+    pl_buf buf;
+    struct pl_plane_data data[4];
+    struct pl_avalloc *alloc = malloc(sizeof(struct pl_avalloc));
+    int planes = pl_plane_data_from_pixfmt(data, NULL, pic->format);
+    if (!(avctx->codec->capabilities & AV_CODEC_CAP_DR1) || !planes)
+        goto fallback;
+    if (!gpu || !gpu->limits.thread_safe || !gpu->limits.max_mapped_size)
+        goto fallback;
+
+    memset(pic->data, 0, sizeof(pic->data));
+    memset(pic->linesize, 0, sizeof(pic->linesize));
+    memset(pic->buf, 0, sizeof(pic->buf));
+    pic->extended_data = pic->data;
+    pic->extended_buf = NULL;
+
+    avcodec_align_dimensions2(avctx, &width, &height, linesize_align);
+    if ((ret = av_image_fill_linesizes(pic->linesize, pic->format, width)))
+        return ret;
+
+    for (int p = 0; p < planes; p++) {
+        int align = PL_MAX(linesize_align[p], gpu->limits.align_tex_xfer_stride);
+        pic->linesize[p] = PL_ALIGN2(pic->linesize[p], align);
+        pic->linesize[p] = PL_ALIGN(pic->linesize[p], data[p].pixel_stride);
+    }
+
+    // Properly align each plane offset, using the pointers as a proxy to
+    // figure out the size of each plane
+    ret = av_image_fill_pointers(pointers, pic->format, height, NULL, pic->linesize);
+    if (ret < 0) {
+        free(alloc);
+        return ret;
+    }
+    pointers[planes] = (uint8_t *) (uintptr_t) ret;
+
+    for (int p = 0; p < planes; p++) {
+        size_t plane_size = (uintptr_t) pointers[p+1] - (uintptr_t) pointers[p];
+        offsets[p] = PL_ALIGN2(total_size, gpu->limits.align_tex_xfer_offset);
+        offsets[p] = PL_ALIGN(offsets[p], data[p].pixel_stride);
+        total_size = offsets[p] + plane_size;
+    }
+
+    if (total_size > gpu->limits.max_mapped_size)
+        goto fallback;
+
+    // Create data buffer
+    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .size = total_size,
+        .memory_type = PL_BUF_MEM_HOST,
+        .host_mapped = true,
+    });
+    if (!buf)
+        goto fallback;
+
+    for (int p = 0; p < planes; p++)
+        pic->data[p] = buf->data + offsets[p];
+
+    // Create buffer ref
+    pic->buf[0] = av_buffer_create(buf->data, total_size, pl_avalloc_free, alloc, 0);
+    *alloc = (struct pl_avalloc) {
+        .magic = { PL_MAGIC0, PL_MAGIC1 },
+        .gpu = gpu,
+        .buf = buf,
+    };
+
+    if (!pic->buf[0]) {
+        pl_buf_destroy(gpu, &buf);
+        free(alloc);
+        return AVERROR(ENOMEM);
+    }
+
+    return 0;
+
+fallback:
+    free(alloc);
+    return avcodec_default_get_buffer2(avctx, pic, flags);
+}
+
+#undef PL_MAGIC0
+#undef PL_MAGIC1
+#undef PL_ALIGN
+#undef PL_ALIGN2
+#undef PL_MAX
+
 #endif // LIBPLACEBO_LIBAV_H_
